/* ParSGCT code release source file.                                              Copyright (c) 2015-2016 Peter Edward Strazdins. All rights reserved.
   Licensed under the terms of the BSD License as described n the COPYING file.
   This comment must be retained in any redistributions of this source file.
*/
// Parallel 3D Advection program
// written by Peter Strazdins, Jun 14

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h> //getopt(), gethostname()
#include <time.h>
#include <mpi.h>
#include <assert.h>
#ifdef _OPENMP
#include <omp.h>
#endif

#include "Advect3D.h"
#include "LinGrid3D.h"
#include "GridCombine3D.h"
#include "HierachDecomp3D.h"
#include "SparseGrid3D.h"

//#define CHECK_SINGLE // use to send scatter via single grids

#define USAGE   "threeDimAdvect [-d] [-2] [-f] [-v v] [-V vT] [-m method] [-l level][-h] [-p pD1] [-q pD2] [-r dP3] [-P P] [-s sgP] [-T tF] [-S steps] [-C nC] [-n] [g.x [g.y [g.z]]]] \n\t "
#define CONSTRAINTS "l <= g.x,g.y,g.z; 2^sgP <= nprocs; sgP<=g.x+g.y+g.z-dim(l-1), where dim = 2 or 3\n\t "
#define DEFAULTS "l=1 v=vT=0 m=2 (2nd order), g.x=g.y=g.z=6 2^sgP~=nprocs pD2=(pD1+1)/2 pD1=(pD0+1)/2 dP0=1 P=1 tF=0.25 nC=1 B=1"
#define NOTES "-2 for 2D (forces gz=0); -F for full combined grid; use -n to print required nprocs only"
#define OPTCHARS "v:m:l:s:dhp:q:r:P:T:S:C:V:n2b:F"

#define NOT_SET -1

/*  0: top-level messages from rank 0, 1: 1-off messages from all ranks,
    2: per-iteration messages, 3: dump data: one-off, 4: dump data, per itn.
 */
static int verbosity = 0;    // v above    
static int verbTim  = -1;     // vT above. Verbosity of timer output 
static bool is2D = false;    // if set, perform 2D simulation (force g.z==0)
static bool dbgGridComb = false;  // set to test/debug combination alg
static int method = NOT_SET;  
static int level = 1;        // use combination algorithm for l>1 
static Vec3D<int> grid;
static int sgProcs = NOT_SET; //2^sgProcs used for sparse grid collection
static double tF = 0.25,     // final time; must it be a multiple of 0.25?
	      CFL = 0.25;    // CFL condition number 
static int rank, nprocs;     // MPI values
static int pD[3];            // number of processes for outer/inner/... planes
static int P = 1;            // number of OpenMP threads per process
static int steps = NOT_SET;  // #timesteps for sparse grid (dept. on grid & TF)
static int nC = 1;           // apply combination algorithm nC times over TF
static bool fixedDt = true;  // if set, use a fixed dt across grids
static bool fixedProcs = true; // signifies #processes is fixed across diagonal
                             // Otherwise pD refers to the central points and 
                             // the number doubles as you go outwards.
                             // If set, only gets good load balance if fixedDt
static int B = 1;            // block size - only for debug mode
static bool hierarchize = false; // is hierarchical basis SGCT
static int useFullGrid = false; // if 1, use a dense grid data struct for s.g.

#define IS_POWER2(n) (((n) & (n-1)) == 0)

// print a usage message for this program and exit with a status of 1
void usage(std::string msg) {
  if (rank==0) {
    printf("threeDimAdvect: %s\n", msg.c_str());
    printf("usage: %s\n\tconstraints: %s\n\tdefault values: %s\n\tnotes: %s", 
	   USAGE, CONSTRAINTS, DEFAULTS, NOTES);
    fflush(stdout);
  }
  exit(1);
}

void getArgs(int argc, char *argv[]) {
  extern char *optarg; // points to option argument (for -p option)
  extern int optind;   // index of last option parsed by getopt()
  extern int opterr;
  char optchar;        // option character returned my getopt()
  opterr = 0;          // suppress getopt() error message for invalid option
  pD[0] = pD[1] = pD[2] = 0;
  bool optNgiven = false; //records  if -n was given 

  while ((optchar = getopt(argc, argv, OPTCHARS)) != -1) {
    // extract next option from the command line     
    switch (optchar) {
    case 'v':
      if (sscanf(optarg, "%d", &verbosity) != 1) // invalid integer 
	usage("bad value for verbose");
      break;
    case 'm':
      if (sscanf(optarg, "%d", &method) != 1) // invalid integer 
	usage("bad value for method");
      break;
    case 'l':
      if (sscanf(optarg, "%d", &level) != 1) // invalid integer 
	usage("bad value for level");
      break;
    case 's':
      if (sscanf(optarg, "%d", &sgProcs) != 1) 
	usage("bad value for sgProcs");
      break;
    case 'p':
      if (sscanf(optarg, "%d", &pD[0]) != 1) 
	usage("bad value for pD0");
      break;
    case 'q':
      if (sscanf(optarg, "%d", &pD[1]) != 1) 
	usage("bad value for pD1");
      break;
    case 'r':
      if (sscanf(optarg, "%d", &pD[2]) != 1) 
	usage("bad value for pD2");
      break;
    case 'P':
      if (sscanf(optarg, "%d", &P) != 1) 
	usage("bad value for P");
      break;
    case 'd':
      dbgGridComb = true;
      break;
    case 'h':
      hierarchize = true;
      break;
    case '2':
      is2D = true;
      break;
    case 'S':
      if (sscanf(optarg, "%d", &steps) != 1) 
	usage("bad value for steps");
      break;
    case 'C':
      if (sscanf(optarg, "%d", &nC) != 1) 
	usage("bad value for nC");
      break;
    case 'T':
      if (sscanf(optarg, "%lf", &tF) != 1) 
	usage("bad value for tF");
      break;
    case 'V':
      if (sscanf(optarg, "%d", &verbTim) != 1) 
	usage("bad value for vT");
      break;
    case 'F':
      useFullGrid = true;
      break;
    case 'n':
      optNgiven = true;
      break;
    case 'b':
      if (sscanf(optarg, "%d", &B) != 1)
        usage("bad value for B");
      break;
    default:
      usage("unknown option");
      break;
    } //switch 
   } //while

  grid.x = grid.y = grid.z = 6;
  if (optind < argc)
    if (sscanf(argv[optind], "%d", &grid.x) != 1) 
      usage("bad value g.x");
  if (optind+1 < argc)
    if (sscanf(argv[optind+1], "%d", &grid.y) != 1) 
      usage("bad value g.y");
  if (optind+2 < argc)
    if (sscanf(argv[optind+2], "%d", &grid.z) != 1) 
      usage("bad value g.z");
  if (is2D)
    grid.z = 0;

  if (method == NOT_SET)
    method = Advect3D::bestMethod(is2D);
    
  if (sgProcs == NOT_SET) // use as many process as possible for sparse grid
    sgProcs = (int) log2((double)nprocs);

  fixedProcs = fixedDt;
  if (pD[0] == 0)
      pD[0] = 1;
      // pD[0] = GridCombine3D::getPD0(nprocs, level, fixedProcs); 
  if (pD[1] == 0)
    pD[1] = GridCombine3D::getPD1(pD[0], fixedProcs);
  if (pD[2] == 0)
    pD[2] = GridCombine3D::getPD1(pD[1], fixedProcs);

  int dx = 1 << std::max(grid.x,  std::max(grid.y, grid.z));
  if (steps == NOT_SET) 
    steps = (int) (tF/CFL * dx);
  else
    tF = CFL * steps / (1.0*dx); 

  if (rank==0 && optNgiven) {    
    printf("%d MPI processes are needed for given command line parameters\n",
	   GridCombine3D::nProcs(is2D, level, pD, fixedProcs));
    exit(0);
  }

  if (!dbgGridComb && B > 1) 
    usage("B > 1 only available in debug mode\n");

#ifdef _OPENMP
  omp_set_num_threads(P);
#endif
  if (rank==0) {
    printf("Level %d combination alg (hier=%d) on grid " V3DFMT " to be applied %d times with %d,%d,%d procs/grid %s, %s combine on 2^%d procs (%d MPI procs) B=%d\n", 
	   level, (int) hierarchize, V3DLST(grid), nC, pD[0], pD[1], pD[2],
	   fixedProcs? "": "(doubling outwards from center)", 
	   useFullGrid? "f.g.": "s.g.", sgProcs, nprocs, B); 
#ifdef _OPENMP
    if (omp_get_max_threads() > 1  &&  verbosity > 0) {
      printf("\t%d threads per proc\n", omp_get_max_threads());
#pragma omp parallel
     {
       printf("%d: hello from thread %d\n", rank, omp_get_thread_num());
      }
#endif
    }
   }

  if (level > std::min(grid.x,  std::min(grid.y, is2D? grid.y: grid.z)))
    usage("level must be <= min(g)");

  if ((1 << sgProcs) > nprocs)
    usage("sgProcs too large for required nprocs" );
  if (sgProcs > grid.x+grid.y+grid.z - (is2D? 2: 3)*(level-1))
    usage("sgProcs too large for given g and level");
  if (nprocs != GridCombine3D::nProcs(is2D, level, pD, fixedProcs))
    usage("nprocs fails its constraint");      
  if (hierarchize && !(IS_POWER2(pD[0]) && IS_POWER2(pD[1]) && IS_POWER2(pD[2])))
    usage("hierarch alg available only for power of 2 process grids");
  if (B > 1 && hierarchize)
    usage("hierarch alg available so far only for B=1");


  if (hierarchize)
    useFullGrid = true;
} //getArgs()

static void printLocGlobAvgs(std::string name, double total, int nlVals, 
			     bool is2d, Vec3D<int> gix, MPI_Comm comm) {
  int ngVals = GridCombine3D::gridSz_(is2d, gix).prod();
  double v[1];  
  if (verbosity > 0)  
    printf("%d: grid (%d,%d,%d): local avg %s is %.2e\n", rank, gix.x, gix.y, 
	   gix.z, name.c_str(), nlVals==0? 0.0: total / nlVals);
  MPI_Reduce(&total, v, 1, MPI_DOUBLE, MPI_SUM, 0, comm);
  int grank;
  MPI_Comm_rank(comm, &grank);
  if (grank == 0)
    printf("%d: grid (%d,%d,%d): avg %s %.2e\n", rank, gix.x, gix.y, gix.z, 
	   (char *) name.c_str(), v[0] / ngVals);
}

int main(int argc, char** argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  getArgs(argc, argv);

  Timer * timer = new Timer;

  GridCombine3D * gc = 
    new GridCombine3D(level, pD, fixedProcs, sgProcs, is2D,
		      grid, timer, MPI_COMM_WORLD, verbosity);
  HierachDecomp3D * hgc = new HierachDecomp3D (gc, verbosity);
  
  MPI_Comm myGridComm;   int grank;
  assert (gc->getGid(rank) == gc->gidU);  
  MPI_Comm_split(MPI_COMM_WORLD, gc->getGid(rank), gc->getGrank(rank), 
		 &myGridComm);
  MPI_Comm_rank(myGridComm, &grank);

  ProcGrid3D *g = gc->pgU, *sg = gc->pgS;
  Vec3D<int> myGrid = gc->gxU; int mygid = gc->gidU;
  HaloArray3D *u = new HaloArray3D(g->G2L(gc->gridSz(myGrid)), 
				   Vec3D<int>(1,1,!gc->is2D()), B);
  SparseGrid3D *usg = 0;
  if (!hierarchize) 
    usg = new SparseGrid3D(useFullGrid, level, is2D, grid, sg, B);
 
  Advect3D *adv = new Advect3D(gc->gridSz(myGrid), Vec3D<double>(1.0), 
			       tF, CFL, method, verbosity, timer);
  Advect3D *advsg = new Advect3D(gc->gridSz(grid), Vec3D<double>(1.0), 
				 tF, CFL, method, verbosity, timer);
  if (fixedDt) 
    // sparse gid will also same dt as that needed on grids 
    // (level,1) and (1, level), the finest dt
    adv->dt = advsg->dt;

  LinGrid3D *ulg = new LinGrid3D(gc->gridSz(myGrid), 
				 gc->gridSz1(grid) / gc->gridSz1(myGrid), B);
  LinGrid3D *slg = new LinGrid3D(gc->gridSz(grid), Vec3D<int>(1), B); 
  if (dbgGridComb) // use 1.0 on each grid
    gc->overrideCombCoeff(1.0);    

  if (rank == 0) {
    if (not dbgGridComb)
      printf("%dD advection via method %d on grid %d,%d,%d for time %e = max %d steps (dt=%e -%sfixed)\n", 
	     is2D? 2: 3, method, grid.x, grid.y, grid.z, tF, steps, 
	     adv->dt, fixedDt? "": "not "); 
    else
      printf("testing grid combination on grid %d,%d,%d for %d reps\n",
	     grid.x, grid.y, grid.z, nC);
  }
  if (verbosity > 0) { 
    char hostName[128];
    gethostname(hostName, sizeof(hostName));
    printf("%d: process (%d,%d,%d) on d=%d grid %d=(%d,%d,%d) with %dx%dx%d processes", 
	   rank, g->id.x, g->id.y, g->id.z, gc->diagRank(gc->gidU), mygid, 
	   myGrid.x, myGrid.y, myGrid.z, g->P.x, g->P.y, g->P.z);
    printf(" computes %dx%dx%d points from (%d,%d,%d) on host %s\n", 
	   u->l.x, u->l.y, u->l.z, g->L2G0(0, adv->gridSize.x), 
	   g->L2G0(1, adv->gridSize.y), g->L2G0(2, adv->gridSize.z), hostName);
  }
  if (rank == 0)
    printf("process (%d,%d,%d) of sparse grid (%d,%d,%d) on %dx%dx%d procs\n",
	   sg->id.x, sg->id.y, sg->id.z, grid.x, grid.y, grid.z,
	   sg->P.x, sg->P.y, sg->P.z);

  int nPtsL = u->l.prod();
  if (not dbgGridComb) {
    adv->initGrid(u, g);
    adv->updateBoundary(u, g);
    // printLocGlobAvgs("norm of initial field", u->norm1(), nPtsL, is2D, myGrid,
    //		     myGridComm);
  } else     
    ulg->initGrid(u, g);

  if (verbosity > 3)
    u->print(rank, "initial field");
  if (!hierarchize)
    usg->zero(); // note: gatherRecv() does not zero it 

  double t = 0.0; 
  double tC = 0.0; // time at last combine operation
  for (int s = 0; s < nC; s++) {  
    if (not dbgGridComb) {
      double dtC = tF / nC;
      t += adv->simulateAdvection(u, g, dtC);
      if (verbosity > 3)
	u->print(rank, "updated field");
      if (nC == 1)
        printLocGlobAvgs("error of updated field", adv->checkError(t, u, g),
			 nPtsL, is2D, myGrid, myGridComm);
    }
    tC = t;
#ifndef CHECK_SINGLE
    if (!hierarchize)
      gc->gatherScatter(u, usg);
    else
      hgc->gatherScatter(u);
#else /*test single send at a time*/
    assert (!hierarchize);
    for (int g=0; g < gc->nGrids(); g++) {
      gc->singleGridCombine(g);
      gc->gather(u, usg);
    }
    gc->singleGridCombine(GridCombine3D::GID_ALL);
    gc->scatter(u, usg);
#endif
  } // for (s...)
  if (not dbgGridComb) 
    t += adv->simulateAdvection(u, g, tF-t);

  // in debug mode, final grid should be scaleFactor * its initial value
  double scaleFactor = std::pow(double(gc->nGrids()), nC);
  if (nC > 0 && !hierarchize) {
    double sgError = 0.0;
    if (sg->myrank >= 0) // this process holds part of the sparse grid
	sgError = dbgGridComb? slg->checkError(scaleFactor, usg, sg):
	                       advsg->checkError(tC, usg, sg);
    printLocGlobAvgs("error of combined field", sgError, usg->numElts(), 
		     is2D, grid, MPI_COMM_WORLD);
    if (verbosity > 3 && sg->myrank >= 0)
      usg->print(rank, "combined field");
  }
  
#if 0
  if (grank == 0 and not dbgGridComb) 
    printf("%d: grid (%d,%d,%d): final time: %e (target %e)\n", rank, 
	   myGrid.x, myGrid.y, myGrid.z, t, tF);
#endif
  double uError = dbgGridComb? ulg->checkError(scaleFactor, u, g):
                               adv->checkError(t, u, g);                 
  printLocGlobAvgs("error of final field", uError, 	
		   nPtsL, is2D, myGrid, myGridComm);  
  if (verbosity > 3)
    u->print(rank, "final field");
  
  timer->dump(myGridComm, verbTim);

  delete gc; delete adv; delete advsg; 
  delete u; 
  if (!hierarchize)
    delete usg;
  delete ulg; delete slg; delete timer;

  MPI_Finalize();
  return 0;
} //main()

